`timescale 1ns / 10ps

module clock_setup(init_clk_p, init_clk_n, setup_scl, setup_sda, i2c_error_led,
  refclk_locked, clkgen_rst_n, i2c_mux_rst_n, init_clk);

  input  init_clk_p;
  input  init_clk_n;
  output  i2c_error_led;
  output  refclk_locked;
  output  clkgen_rst_n;
  output  i2c_mux_rst_n;
  output  init_clk;
  inout  setup_scl;
  inout  setup_sda;
  reg  refclk_locked;
  reg  reset;
  reg [12:0] div_counter;
  reg  sclk_logic;
  reg  sdata_logic;
  reg  sdata_sample;
  reg  i2c_en;
  reg  pre_en;
  reg [3:0] state;
  reg  first;
  reg  dir_write;
  reg  save_direction;
  reg [7:0] write_byte;
  reg [7:0] read_byte;
  reg [2:0] idx;
  reg [5:0] ctrl_idx;
  reg [1:0] stopmode;
  reg [9:0] ctrlword;
  reg  i2c_error;
  reg  kickoff;
  reg [10:0] wakeup_cnt;
  reg [13:0] led_cnt;
  wire  init_clk_i;
  wire [7:0] clkgen_addr;
  wire [7:0] clkgen_read_addr;

   parameter clkfreq = 125;

   parameter st_idle = 0,
	     st_start = 1,
	     st_fetch = 2,
	     st_bit0 = 3,
	     st_bit1 = 4,
	     st_bit2 = 5,
	     st_ack0 = 6,
	     st_ack1 = 7,
	     st_stop0 = 8,
	     st_stop1 = 9,
	     st_stop2 = 10; // Represented by "default"

   initial wakeup_cnt = 0;
   initial i2c_error = 0;
   initial refclk_locked = 0;
   initial reset = 1;

   IBUFDS ibufds (.O(init_clk_i), .I(init_clk_p), .IB(init_clk_n));
   BUFG bufg (.O(init_clk), .I(init_clk_i));

   assign clkgen_addr = 208;
   assign clkgen_read_addr = clkgen_addr | 1;

   assign clkgen_rst_n = !reset;
   assign i2c_mux_rst_n = !reset;

   assign i2c_error_led = led_cnt[13];

   // Emulated open collector output
   // Note that sclk and sdata must be pulled up, possibly with
   // a PULLUP constraint on the IOB (or a 10 kOhm ext. resistor)

   assign setup_scl = sclk_logic ? 1'bz : 1'b0 ;
   assign setup_sda = sdata_logic ? 1'bz : 1'b0 ;

   always @(posedge init_clk)
     begin
	if (i2c_en && !(&wakeup_cnt))
	  wakeup_cnt <= wakeup_cnt + 1;

	reset <= (wakeup_cnt < 16);
	kickoff <= (wakeup_cnt == 2016); // 10 ms after reset

	if (!i2c_error)
	  led_cnt <= 0;
	else if (i2c_en)
	  led_cnt <= led_cnt + 1;

	i2c_en <= pre_en;
	sdata_sample <= setup_sda;

	if (div_counter == ((clkfreq * 5) - 1))
	  begin
	     div_counter <= 0;
	     pre_en <= 1;
	  end
	else
	  begin
	     div_counter <= div_counter + 1;
	     pre_en <= 0;
	  end

	case (ctrl_idx)
	  0: ctrlword <= { 2'd0, 8'he8 };
	  1: ctrlword <= { 2'd1, 8'h10 };
	  2: ctrlword <= { 2'd0, clkgen_addr };
	  3: ctrlword <= { 2'd0, 8'h00 };
	  4: ctrlword <= { 2'd0, 8'h54 };
	  5: ctrlword <= { 2'd0, 8'he4 };
	  6: ctrlword <= { 2'd0, 8'h32 };
	  7: ctrlword <= { 2'd0, 8'h55 };
	  8: ctrlword <= { 2'd0, 8'h12 };
	  9: ctrlword <= { 2'd0, 8'hed };
	  10: ctrlword <= { 2'd0, 8'h3f };
	  11: ctrlword <= { 2'd0, 8'h28 };
	  12: ctrlword <= { 2'd0, 8'h00 };
	  13: ctrlword <= { 2'd0, 8'hc0 };
	  14: ctrlword <= { 2'd0, 8'h00 };
	  15: ctrlword <= { 2'd1, 8'h40 };
	  16: ctrlword <= { 2'd0, clkgen_addr };
	  17: ctrlword <= { 2'd0, 8'h13 };
	  18: ctrlword <= { 2'd0, 8'h29 };
	  19: ctrlword <= { 2'd0, 8'h35 };
	  20: ctrlword <= { 2'd0, 8'hfc };
	  21: ctrlword <= { 2'd0, 8'hdf };
	  22: ctrlword <= { 2'd0, 8'h18 };
	  23: ctrlword <= { 2'd0, 8'h39 };
	  24: ctrlword <= { 2'd1, 8'h60 };
	  25: ctrlword <= { 2'd0, clkgen_addr };
	  26: ctrlword <= { 2'd0, 8'h1f };
	  27: ctrlword <= { 2'd0, 8'h00 };
	  28: ctrlword <= { 2'd0, 8'h00 };
	  29: ctrlword <= { 2'd0, 8'h05 };
	  30: ctrlword <= { 2'd0, 8'h00 };
	  31: ctrlword <= { 2'd0, 8'h00 };
	  32: ctrlword <= { 2'd1, 8'h05 };
	  33: ctrlword <= { 2'd0, clkgen_addr };
	  34: ctrlword <= { 2'd0, 8'h28 };
	  35: ctrlword <= { 2'd0, 8'hc0 };
	  36: ctrlword <= { 2'd0, 8'h88 };
	  37: ctrlword <= { 2'd0, 8'hb7 };
	  38: ctrlword <= { 2'd0, 8'h00 };
	  39: ctrlword <= { 2'd0, 8'h1d };
	  40: ctrlword <= { 2'd0, 8'hc2 };
	  41: ctrlword <= { 2'd0, 8'h00 };
	  42: ctrlword <= { 2'd0, 8'h1d };
	  43: ctrlword <= { 2'd1, 8'hc2 };
	  44: ctrlword <= { 2'd0, clkgen_addr };
	  45: ctrlword <= { 2'd0, 8'h37 };
	  46: ctrlword <= { 2'd1, 8'h1b };
	  47: ctrlword <= { 2'd0, clkgen_addr };
	  48: ctrlword <= { 2'd0, 8'h89 };
	  49: ctrlword <= { 2'd0, 8'h01 };
	  50: ctrlword <= { 2'd0, 8'h0c };
	  51: ctrlword <= { 2'd1, 8'hcc };
	  52: ctrlword <= { 2'd0, clkgen_addr };
	  53: ctrlword <= { 2'd0, 8'h8e };
	  54: ctrlword <= { 2'd0, 8'h00 };
	  55: ctrlword <= { 2'd1, 8'h00 };
	  56: ctrlword <= { 2'd0, clkgen_addr };
	  57: ctrlword <= { 2'd0, 8'h88 };
	  58: ctrlword <= { 2'd1, 8'h40 };
	  59: ctrlword <= { 2'd0, clkgen_addr };
	  60: ctrlword <= { 2'd2, 8'h82 };
	  61: ctrlword <= { 2'd0, clkgen_read_addr };
	  62: ctrlword <= { 2'd1, 8'hff };
	  default: ctrlword <= 0; // Should never reach this
	endcase

	// State machine

 	if (i2c_en)
	  case (state)
	    st_idle:
	      begin
		 sclk_logic <= 1;
		 sdata_logic <= 1;

		 if (kickoff || (ctrl_idx != 0))
		   state <= st_start;
	      end

	    st_start:
	      begin
		 sdata_logic <= 0; // Start condition
		 first <= 1;
		 dir_write <= 1;

		 state <= st_fetch;
	      end

	    st_fetch:
	      begin
		 sclk_logic <= 0;
		 idx <= 7;
		 { stopmode, write_byte } <= ctrlword;
		 state <= st_bit0;
	      end

	    st_bit0:
	      begin
		 if (dir_write)
		   sdata_logic <= write_byte[idx];
		 else
		   sdata_logic <= 1;

		 state <= st_bit1;
	      end

	    st_bit1:
	      begin
		 sclk_logic <= 1;
		 read_byte[idx] <= sdata_sample;

		 state <= st_bit2;
	      end

	    st_bit2:
	      begin
		 sclk_logic <= 0;

		 idx <= idx - 1;

		 if (idx != 0)
		   state <= st_bit0;
		 else
		   state <= st_ack0;
	      end

	    st_ack0:
	      begin
		 if (dir_write)
		   sdata_logic <= 1; // The slave should ack
		 else
		   begin
		      sdata_logic <= 0; // We ack on read
		      refclk_locked <= !read_byte[0]; // !LOL_INT
		   end

		 save_direction <= !write_byte[0];
		 state <= st_ack1;
	      end

	    st_ack1:
	      begin
		 sclk_logic <= 1;

		 case (stopmode)
		   1: // Stop after this
		     if (dir_write)
		       state <= st_stop0;
		     else
		       state <= st_stop2; // Don't toggle clock in read mode

		   2: // Restart after this. Intended for write mode only
		     state <= st_stop0;

		   default: // stopmode == 0, go on to next word
		     begin
			if (first)
			  dir_write <= save_direction;

			first <= 0;

			state <= st_fetch;
		     end
		 endcase

		 if (dir_write && sdata_sample) // NAK, stop immediately
		   begin
		      stopmode <= 1;
		      state <= st_stop0;
		      ctrl_idx <= 0;
		      i2c_error <= 1;
		   end

		 if (ctrl_idx != 62)
		   ctrl_idx <= ctrl_idx + 1;
		 else if (!refclk_locked)
		   ctrl_idx <= 59; // Continue loop on reading register
		 else
		   ctrl_idx <= 0;
	      end

	    // The three stop states toggle the clock once, so that
	    // we're sure that the slave has released the bus, leaving
	    // its acknowledge state. Used only in write direction.

	    st_stop0:
	      begin
		 sclk_logic <= 0;
		 state <= st_stop1;
	      end

	    st_stop1:
	      begin
		 // If stopmode == 2, don't create a stop condition, but
		 // do toggle the clock.
		 sdata_logic <= (stopmode == 2) ? 1 : 0;
		 state <= st_stop2;
	      end

	    default: // Normally this is st_stop2
	      begin
		 sclk_logic <= 1;

		 // st_idle will raise sdata to '1', making a stop condition
		 state <= st_idle;
	      end
	  endcase

	if (i2c_error)
	  refclk_locked <= 0; // Just to be really sure about this

	if (reset) // Override all above.
	  begin
	     state <= st_idle;
	     ctrl_idx <= 0;
	     i2c_error <= 0;
	     refclk_locked <= 0;
	  end
     end

endmodule
