`timescale 1ns / 10ps

module ultrascale_frontend(gtx_rxn, gtx_rxp, gtx_txn, gtx_txp, gtx_refclk_n,
  gtx_refclk_p, pipe_rx_clk, pipe_clk, pll_locked, frontend_rst, refclk_locked,
  mgt_en, lfps_en, mgt_powerdown, rx_dontmess, rx_phy_ready, rx_resync,
  rx_detect_revpolarity, rx_detect_revpolarity_clear, rx_align_enable,
  rx_elecidle, receiver_detect, receiver_present, receiver_present_valid,
  pipe_rx, pipe_rx_k, pipe_rx_valid, pipe_tx, pipe_tx_k, pipe_tx_ready,
  init_clk, version_ultrascale_1_0);

  input  gtx_rxn;
  input  gtx_rxp;
  input  gtx_refclk_n;
  input  gtx_refclk_p;
  input  refclk_locked;
  input  mgt_en;
  input  lfps_en;
  input  mgt_powerdown;
  input  rx_dontmess;
  input  rx_resync;
  input  rx_detect_revpolarity;
  input  rx_align_enable;
  input  receiver_detect;
  input [31:0] pipe_tx;
  input [3:0] pipe_tx_k;
  input  init_clk;
  output  gtx_txn;
  output  gtx_txp;
  output  pipe_rx_clk;
  output  pipe_clk;
  output  pll_locked;
  output  frontend_rst;
  output  rx_phy_ready;
  output  rx_detect_revpolarity_clear;
  output  rx_elecidle;
  output  receiver_present;
  output  receiver_present_valid;
  output [31:0] pipe_rx;
  output [3:0] pipe_rx_k;
  output  pipe_rx_valid;
  output  pipe_tx_ready;
  output  version_ultrascale_1_0;
  reg  pll_locked;
  reg  rx_phy_ready;
  reg  rx_detect_revpolarity_clear;
  reg  rx_elecidle;
  reg  receiver_present;
  reg  receiver_present_valid;
  reg [2:0] mgt_reset_rx;
  reg [2:0] mgt_reset_tx;
  reg  txelecidle;
  reg [4:0] wait_rx_phy_ready;
  reg  rxelecidle_d;
  reg  rxelecidle_d2;
  reg [4:0] elecidle_cnt;
  reg [31:0] txdata;
  reg [3:0] txdata_k;
  reg  txdata_extend;
  reg [1:0] lfps_toggler;
  reg  tx_coding_bypass;
  reg [5:0] wakeup_counter;
  reg  awake;
  reg  wait_for_wakeup;
  reg  wait_for_wakeup_d;
  reg  wait_for_wakeup_sync;
  reg  cpll_lock_d;
  reg  cpll_lock_d2;
  reg  cpll_lock_rx_d;
  reg  cpll_lock_rx_d2;
  reg  gtwiz_reset_all;
  reg  gtpowergood_d;
  reg  gtpowergood_d2;
  reg  do_wakeup;
  reg  reset_tx_done_d;
  reg  reset_tx_done_d2;
  reg  reset_rx_done_d;
  reg  reset_rx_done_d2;
  reg  rxresetdone_d;
  reg  rxresetdone_sync;
  reg  phalign_done_d;
  reg  phalign_done_sync;
  reg [1:0] powerdown;
  reg  mgt_powerdown_d;
  reg  txdetectrx;
  reg [6:0] polarity_debounce_cnt;
  reg  rxpolarity;
  reg  rx_align_enable_d;
  reg  rx_align_enable_d2;
  wire  reset_rx;
  wire  reset_tx;
  wire  rxelecidle_i;
  wire  gt_refclk;
  wire  cpll_lock;
  wire  gtpowergood;
  wire  reset_tx_done;
  wire  reset_rx_done;
  wire  phalign_done;
  wire  phalign_error;
  wire  phystatus;
  wire [2:0] rxstatus;

   parameter RXELECIDLE_DEBOUNCE = 14; // ~ f_usrclk * 0.1 us

   initial wakeup_counter = 0;
   initial gtwiz_reset_all = 0;
   initial awake = 0;

   assign version_ultrascale_1_0 = 1;

   assign pipe_tx_ready = 1;
   assign pipe_rx_valid = 1;

   assign reset_rx = mgt_reset_rx[2];
   assign reset_tx = mgt_reset_tx[2];

   assign frontend_rst = !awake;

   always @(posedge init_clk)
     begin
	gtpowergood_d <= gtpowergood;
	gtpowergood_d2 <= gtpowergood_d;

	reset_tx_done_d <= reset_tx_done;
	reset_tx_done_d2 <= reset_tx_done_d;

	reset_rx_done_d <= reset_rx_done;
	reset_rx_done_d2 <= reset_rx_done_d;

	do_wakeup <= gtpowergood_d2 && refclk_locked;

	if (!do_wakeup)
	  wakeup_counter <= 0;
	else if (!(&wakeup_counter))
	  wakeup_counter <= wakeup_counter + 1;

	// The GT's reset input is asynchronous, one clock cycle is enough
	gtwiz_reset_all <= (wakeup_counter == 8);

	// The purpose of wakeup_counter is to ensure that the reset_*_done
	// signals have gone low after gtwiz_reset_all was asserted
	wait_for_wakeup <= !(reset_tx_done_d2 && reset_rx_done_d2 &&
			     &wakeup_counter) && !awake;

	if (!do_wakeup)
	  awake <= 0;
	else if (!wait_for_wakeup)
	  awake <= 1;
     end

   always @(posedge pipe_clk)
     begin
	// Polarity detection debouncing is required to clear the pipeline
	// of data before the next verdict.

	rx_detect_revpolarity_clear <= (polarity_debounce_cnt != 0);

	if (polarity_debounce_cnt != 0)
	  polarity_debounce_cnt <= polarity_debounce_cnt - 1;
	else if (rx_detect_revpolarity)
	  begin
	     polarity_debounce_cnt <= polarity_debounce_cnt - 1; // Wrap to max
	     rxpolarity <= !rxpolarity;
	  end

	// CPLL lock indicator is asynchronous
	cpll_lock_d <= cpll_lock;
	cpll_lock_d2 <= cpll_lock_d;
	pll_locked <= cpll_lock_d2;

	// rxelecidle_i is an async signal, hence metastability guarded
	rxelecidle_d <= rxelecidle_i;
	rxelecidle_d2 <= rxelecidle_d;

	if (!rxelecidle_d2)
	  elecidle_cnt <= RXELECIDLE_DEBOUNCE;
	else if (elecidle_cnt != 0)
	  elecidle_cnt <= elecidle_cnt - 1;

	rx_elecidle <= (elecidle_cnt == 0);

	// mgt_reset_* assignments are possibly overridden below
	mgt_reset_rx <= { mgt_reset_rx, 1'b0 };
	mgt_reset_tx <= { mgt_reset_tx, 1'b0 };

	if (rx_resync)
	  mgt_reset_rx <= ~0;

	lfps_toggler <= lfps_toggler + 1;

	powerdown <= mgt_powerdown ? 2'b10 : 2'b00;
	mgt_powerdown_d <= mgt_powerdown;

	if (mgt_powerdown_d && !mgt_powerdown)
	  mgt_reset_tx <= ~0;

	if (mgt_en)
	  begin
	     txelecidle <= 0;
	     txdata <= pipe_tx;
	     txdata_k <= pipe_tx_k;
	     txdata_extend <= 0;
	     tx_coding_bypass <= 0;
	  end
	else if (lfps_en)
	  begin
	     txelecidle <= 0;
	     txdata <= { 32{lfps_toggler[1]} };
	     txdata_extend <= lfps_toggler[1];
	     tx_coding_bypass <= 1;
	  end
	else // Electrical Idle
	  begin
	     txelecidle <= 1;
	  end

	receiver_present_valid <= txdetectrx && phystatus;
	receiver_present <= rxstatus[0];

	// txdetectrx is deasserted on the clock following phystatus assertion,
	// following the PIPE spec, and not one clock later, as shown in e.g.
	// UG576. It can't hurt deasserting it when the result is available,
	// and UG576's timing diagram isn't PIPE compliant.

	if (reset_tx || phystatus)
	  txdetectrx <= 0;
	else if (receiver_detect)
	  txdetectrx <= 1;

	rxresetdone_d <= reset_rx_done;
	rxresetdone_sync <= rxresetdone_d;

	phalign_done_d <= phalign_done;
	phalign_done_sync <= phalign_done_d;

	if (reset_rx || !rxresetdone_sync || !phalign_done_sync)
	  wait_rx_phy_ready <= 0;
	else if (!(&wait_rx_phy_ready))
	  wait_rx_phy_ready <= wait_rx_phy_ready + 1;

	rx_phy_ready <= &wait_rx_phy_ready;

	wait_for_wakeup_d <= wait_for_wakeup;
	wait_for_wakeup_sync <= wait_for_wakeup_d;

	if (wait_for_wakeup_sync) // Don't reset until wakeup is done
	  begin
	     mgt_reset_rx <= 0;
	     mgt_reset_tx <= 0;
	  end
     end

   always @(posedge pipe_rx_clk)
     begin
	rx_align_enable_d <= rx_align_enable;
	rx_align_enable_d2 <= rx_align_enable_d;

	cpll_lock_rx_d <= cpll_lock;
	cpll_lock_rx_d2 <= cpll_lock_rx_d;
     end

   ultragt ultragt_ins
     (
      .gtwiz_userclk_tx_reset_in(!refclk_locked),
      .gtwiz_userclk_tx_srcclk_out(),
      .gtwiz_userclk_tx_usrclk_out(),
      .gtwiz_userclk_tx_usrclk2_out(pipe_clk),
      .gtwiz_userclk_tx_active_out(),
      .gtwiz_userclk_rx_reset_in(!refclk_locked),
      .gtwiz_userclk_rx_srcclk_out(),
      .gtwiz_userclk_rx_usrclk_out(),
      .gtwiz_userclk_rx_usrclk2_out(pipe_rx_clk),
      .gtwiz_userclk_rx_active_out(),
      .gtwiz_buffbypass_rx_reset_in(!cpll_lock_rx_d2),
      .gtwiz_buffbypass_rx_start_user_in(phalign_done && phalign_error),
      .gtwiz_buffbypass_rx_done_out(phalign_done),
      .gtwiz_buffbypass_rx_error_out(phalign_error),
      .gtwiz_reset_clk_freerun_in(init_clk),
      .gtwiz_reset_all_in(gtwiz_reset_all),
      .gtwiz_reset_tx_pll_and_datapath_in(1'b0),
      .gtwiz_reset_tx_datapath_in(reset_tx),
      .gtwiz_reset_rx_pll_and_datapath_in(1'b0),
      .gtwiz_reset_rx_datapath_in(reset_rx),
      .gtwiz_reset_rx_cdr_stable_out(),
      .gtwiz_reset_tx_done_out(reset_tx_done),
      .gtwiz_reset_rx_done_out(reset_rx_done),
      .gtwiz_userdata_tx_in(txdata),
      .gtwiz_userdata_rx_out(pipe_rx),
      .drpclk_in(init_clk),
      .gthrxn_in(gtx_rxn),
      .gthrxp_in(gtx_rxp),
      .gtrefclk0_in(gt_refclk),
      .rx8b10ben_in(1'b1),
      .rxcommadeten_in(1'b1),
      .rxelecidlemode_in(2'd0),
      .rxmcommaalignen_in(rx_align_enable_d2),
      .rxoobreset_in(1'b0),
      .rxpcommaalignen_in(rx_align_enable_d2),
      .rxpd_in(powerdown),
      .rxpolarity_in(rxpolarity),
      .tx8b10bbypass_in({ 4'd0, {4{tx_coding_bypass}}}),
      .tx8b10ben_in(1'b1),
      .txctrl0_in({12'd0, { 4{txdata_extend} } }),
      .txctrl1_in({12'd0, { 4{txdata_extend} } }),
      .txctrl2_in({4'd0, txdata_k}),
      .txdetectrx_in(txdetectrx),
      .txelecidle_in(txelecidle),
      .txpd_in(powerdown),
      .txpdelecidlemode_in(1'b1),
      .txpolarity_in(1'b0),
      .cplllock_out(cpll_lock),
      .gthtxn_out(gtx_txn),
      .gthtxp_out(gtx_txp),
      .gtpowergood_out(gtpowergood),
      .phystatus_out(phystatus),
      .rxbyteisaligned_out(),
      .rxbyterealign_out(),
      .rxcominitdet_out(),
      .rxcommadet_out(),
      .rxcomsasdet_out(),
      .rxcomwakedet_out(),
      .rxctrl0_out(pipe_rx_k),
      .rxctrl1_out(),
      .rxctrl2_out(),
      .rxctrl3_out(),
      .rxelecidle_out(rxelecidle_i),
      .rxpmaresetdone_out(),
      .rxstatus_out(rxstatus),
      .txpmaresetdone_out()
      );

   IBUFDS_GTE3 #(
		 .REFCLK_EN_TX_PATH(1'b0),
		 .REFCLK_HROW_CK_SEL(2'd0),
		 .REFCLK_ICNTL_RX(2'd0)
		 ) ibufds_gt
     (
      .I(gtx_refclk_p),
      .IB(gtx_refclk_n),
      .CEB(1'b0),
      .O(gt_refclk),
      .ODIV2()
      );

endmodule
